"""
Specific to the GAPS experiment
Prediction of the energy resolution with the custom ASIC
from the calibration with discrete preamps.
"""

import numpy as np
from dataclasses import dataclass
from .noisemodel import Constants



@dataclass
class ASICConstants : 
    S_w   : float = 0.54*1e-18 #V2/Hz
    A_f   : float = 7.3*1e-14 #V2
    I_eff : float = 2.5*1e-9 # A
    # RC CR2 coefficients
    F_i   : float = 0.64 
    F_nu  : float = 0.853
    F_nuf : float = 0.543


CONSTANTS  = Constants()
ACONSTANTS = ASICConstants()

def enc2_asic(tau, I_L,  A_f, C=70*1e-12):
    """
    ASIC prediction from fitted noisemodel
    values from the preamp measurement.

    Args:
        tau (iterable) : xs - the shaping/peaking time
        I_L (float)    : fitted leakage current from the preamp 
                         measurement
        A_f (float)    : 

    Keyword Args:
        C (float)      : fitted capacity**2 from the preamp
                         measurementi - looks like this is always
                         70 pF
        
    """
    tau = 1e-6*tau
    C_2 = C**2
    A =  2*CONSTANTS.q*(I_L + ACONSTANTS.I_eff)*tau*ACONSTANTS.F_i
    B =  ACONSTANTS.S_w*(C_2)*ACONSTANTS.F_nu*(1/tau)
    C = 2*np.pi*A_f*ACONSTANTS.F_nuf*(C_2)*np.ones(len(tau))
    enc2 = A+B+C
    print ((A[0],B[0], C[0], 'A,B,C'))
    return np.sqrt(enc2)*2.355*CONSTANTS.eps*(1/CONSTANTS.q)*1e-3
    #return enc2



"""
Helper functions for getting data, plotting etc.

"""

import pylab as p

from dataclasses import dataclass

@dataclass
class Measurment:
    """
    A simple container holding spectral fit results
    """
    ptime         : int
    strip         : str
    mtype         : str
    fwhm          : float
    fwhme         : float
    ptimefromfile : int = -1
    filename      : str = ''
    statistics    : int = -1

#################################################################

def get_data_from_summaryfile(sfile, mtype):
    """
    Extract the data from a summaryfile written by FitXrayData

    Args:
        sfile (file descriptor)  : open summaryfile
        mtype (str)              : descriptive string about the 
                                   measurement type. Can be either
                                   'WF'   - waveform analysis
                                   'CAEN' - online CAEN energy analysis
                                   'MC2'  - Data taken with MC2
    """

    data = []
    for line in sfile:
        if line.startswith('#'):
            continue
        line = line.split()
        m = Measurment(ptime = float(line[2]),\
                       mtype = mtype,\
                       strip = line[1],\
                       fwhm  = float(line[3]),\
                       fwhme = float(line[4]))
        data.append(m)
    return data

########################################################################

def get_data_from_mit_file(sfile):
    """
    Extract MC2 data from ascii file format used by MIT
    This will always be MC2 data.

    Args:
        sfile (file descriptor)  : input ascii file
    """
    data = []
    for line in sfile:
        if line.startswith('#'):
            continue
        line = line.split()
        m = Measurment(ptime = float(1000*float(line[2])),\
                       mtype = 'MC2',\
                       strip = f'strip{line[1]}',\
                       fwhm  = float(line[5]),\
                       fwhme = float(line[6]))
        data.append(m)
    return data

########################################################################

def get_stripname(ch):
    """ 
    Mapping channel number <-> strip name
    
    Args:
        ch (int) : channel number [0-7]

    Returns
        str      : strip name

    """
    channel_names = {\
    0 : 'stripA',\
    1 : 'stripB',\
    2 : 'stripC',\
    3 : 'stripD',\
    4 : 'stripE',\
    5 : 'stripF',\
    6 : 'stripG',\
    7 : 'stripH'}
    return channel_names[ch]


########################################################################


def plot_waveform(ax, times, digi_channels):
    """
    Create a plot for the individual waveform, set labels assume
    times is in microseconds and voltages are mV.

    Args:
        times (ndarray)                : Times in microseconds
        voltages (ndarrya)             : Waveform in milliVolts

    Returns:
        None
    """
    ax.plot(times, digi_channels, lw=0.9, color="r", alpha=0.7, label='waveform')
    return ax





"""
Offline analysis on waveforms obtained with dactylos
The shaping is done with code from 
https://github.com/alowell-ssl/awlpico
"""
import pylab as p
import numpy as np
import concurrent.futures as fut
import uproot as up
import tqdm
import os
import os.path
import tqdm

from copy import copy

import hepbasestack as hep 

from .utils import get_stripname, plot_waveform

# shaping stuff 
#from .shaping import GaussShaper, TrapezoidalFilter
from . import shaping as sh

import dactylos

logger = hep.logger.get_logger(dactylos.LOGLEVEL)

########################################################################

def read_waveform(infile, ch, entrystop=None):
    """
    Return waveform data from a rootfile with uproot

    Args:
        infile (str)    : filename
        ch (int)        : digitizer channel

    Keyword Args:
        entrystop (int) : last entry - if None, read all
    Returns:
        ndarray : numpy array with waveform data. Waveform data is in
                  digitizer channels and of length of the record length
                  as set when configuring the digitizer
    """
    f = up.open(infile)
    data = f.get('ch' + str(ch)).get('waveform').array(entrystop=entrystop)
    # we omit empty events here. This should not happen
    # but might happen during digitizer software debugging
    data = [baseline_correction(k, nsamples=1000)[0] for k in data if len(k) > 0]
    #if sh.WAVEFORMTYPE == 'ARRAY':
    #    data = np.array(data, dtype=np.int16)
    # data is in counts here, so it is still technichally not more precise than
    # an int 16
    if sh.WAVEFORMTYPE == 'ARRAY':
        data = np.asarray(data, dtype=np.int16)
    logger.info(f'Read out {len(data)} events for channel {ch}')    
    return ch, data

########################################################################

def baseline_correction(input_pulses, nsamples=2000):
    """
    Do a very simple baseline subtraction. Assume the first nsamples are
    the baseline, and then average over them and subtract it

    Args:
        input (ndarray)    : input waveform

    Keyword Args:
        nsamples (int)     : number of samples for baseline estimation 
                             at the beginning of the input data

    Returns:
        numpy.ndarray      : waveform with a subtracted baseline

    """

    absolute_zero = (2**14)/2
    baseline = input_pulses[:nsamples].mean()
    #print (baseline) 
    #baseline = input[:nsamples].mean() - zero
    #print (f"Baseline calculation gives us {baseline}")
    #return input - np.ones(len(input))*baseline - absolute_zero*np.ones(len(input)), baseline
    if sh.WAVEFORMTYPE == 'ARRAY':
        data = input_pulses - np.ones(len(input_pulses), dtype=np.int16)*baseline, baseline
    if sh.WAVEFORMTYPE == 'LIST':
        data = list(np.asarray(input_pulses, dtype=np.int16) - np.ones(len(input_pulses), dtype=np.int16)*int(baseline)), baseline
    return data
    #return baseline,(input - baseline)

########################################################################

class WaveformAnalysis(object):
    """
    Allow for the read-in and chaining of multiple files

    """

    def __init__(self, njobs=4,\
                 active_channels=[0,1,2,3,4,5,6,7],\
                 use_simple_trapezoid_shaper=False,\
                 shaper_order=4,\
                 adjust_shaper_order_dynamically=False):
        """
        Prepare the waveform analysis. This is a several step process.
        First, tell it what files to use, then load them and finaly analyze.

        Keyword Args:
            njobs (int)                         : number of cpu jobs to use
            use_simple_trapezoid_shaper (bool)  : instead of the Gaussian shaper, a simple trapezoidal
                                                  filter can be used. 

        Returns:
            order (int)           :  -- unknown shaping parameter
            decay_time (float)    :  The decay time of the pulse. This has to
                                     pretty much mach the decaying tail of the pulse.
                                     Decay time has to be given in seconds.
            dt (float)            :  Digitizer sample size (that is 1/sampling rate) in 
                                     seconds. This value is fix for the CAEN 6725 with 
                                     250 MSamples/s
            shaper_order (int)    :  Order of the Gaussian shaper (4 is default)
            adjust_shaper_order_dynamically (bool) : Switch to a higher shaper order
                                                     for larger peakingtimes automatically
        """
        self.files = []
        self.tpexecutor = fut.ThreadPoolExecutor(max_workers=njobs)
        self.ppexecutor = fut.ProcessPoolExecutor(max_workers=njobs)
        self.active_channels = active_channels
        self.channel_data = dict([(k, np.array([])) for k in self.active_channels])  
        self.order = shaper_order
        self.adjust_shaper_order_dynamically = adjust_shaper_order_dynamically
        self.use_simple_trapezoid_shaper = use_simple_trapezoid_shaper
        self.recordlengths = dict()

    def get_recordlengths(self):
        """
        For debugging purposes. Check the recordlength of the 
        Waveforms and return a set for each channel with the
        recordlengths in the actual waveform data. This might 
        differ from the actual setting for whatever reason. 
        Might be even a bug.

        Returns:
            dict     : ch -> set(recordlength)

        """
        result = dict()
        for ch in self.active_channels:
            rl = np.array([len(k) for k in self.channel_data[ch]])
            if len (set(rl)) != 1:
                raise ValueError(f'Waveforms in the SAME CHANNEL do not have the SAME recordlength! Something is terribly wrong! {set(rl)}')
            result[ch] = rl[0]
            
        self.recordlengths = result
        return result

    def get_eventcounts(self):
        """
        Return the number of events per channel. This only makes sense
        after some data has been via read_waveform

        Return:
            dict     : int->int Number of events per channel
        """
        nevents = dict()
        # readjust active channels
        # if one has 0 events, it is not active
        self.active_channels = []
        for ch in self.channel_data.keys():
            nevents[ch] = len(self.channel_data[ch])
            if nevents[ch] > 0:
                self.active_channels.append(ch)
            else:
                logger.warning("Ch {ch} has no observed events! Deactivating...")
        return nevents

    def __del__(self):
        self.tpexecutor.shutdown()
        self.ppexecutor.shutdown()

    def set_shaping_parameters(self, peaktime,
                               order = 4,\
                               decay_time = 80e-6,\
                               dt = 4e-9 ):
        """
        Set all the ncecessary parameters for Alex' gaussian
        shaper in a coherent way. 
        Typically order and decay_time do not need to be changed.

        Args: 
            peaktime (float)      :  The peaking time, that is the time
                                     where the waveform is sampled with 
                                     the gaussian. Peaktime has to be given in 
                                     seconds.

        Keyword Args:
            order (int)           :  -- unknown shaping parameter
            decay_time (float)    :  The decay time of the pulse. This has to
                                     pretty much mach the decaying tail of the pulse.
                                     Decay time has to be given in seconds.
            dt (float)            :  Digitizer sample size (that is 1/sampling rate) in 
                                     seconds. This value is fix for the CAEN 6725 with 
                                     250 MSamples/s
        """
        # peaktime can be e.g. 4e-6, that is 4 microseconds
        self.peaktime = peaktime
        self.order = order
        self.decay_time = decay_time
        self.dt = dt  

    def set_peakingtime_sequence(self, peakingtimes):
        """
        Set a number of peaking times we want to 
        try on each channel

        Args:
            peakingtimes (list)   : peakingtimes in microsec

        """
        self.peakingtime_sequence = peakingtimes

    def set_infiles(self, infiles):
        """
        Define a list of files to be readout later.

        Args:
            infiles (list)   : root files with waveform data
        
        """
        self.files = infiles

    def read_waveforms(self, entrystop=None):
        """
        Read the waveforms from the files

        Keyword Args:
            entrystop (int)   : if not None, read only entrystop waveforms

        Returns:
            None
        """
        future_to_fname = dict()

        for fname in self.files:
            # read out every file once per channel
            for ch in self.active_channels:
                #read_waveform(fname, ch)
                future_to_fname[self.tpexecutor.submit(read_waveform, fname, ch, entrystop = entrystop)] = fname

        logger.info("Readout jobs submitted..")
        init_ch = [False]*8
        for future in fut.as_completed(future_to_fname):
            ch, data = future.result()    
            if sh.WAVEFORMTYPE == 'ARRAY':
                data = np.array([np.array(k, dtype=np.int16) for k in data])
            if not init_ch[ch]:
                self.channel_data[ch] = data
                init_ch[ch] = True
            else:
                self.channel_data[ch] = np.vstack((self.channel_data[ch], data))
        logger.info("Channel data created")
        return None
        #return channel_data

    def plot_waveforms(self, ch, nwaveforms,\
                       prefix='',\
                       savedir='.'):
        """
        Plot some of the waveforms.

        Args:
            ch (int)              : Plot waveforms of this channel
            nwaveforms (int)      : Plot the first nwaveforms
    
        Keyword Args: 
            prefix (str)          : A prefix to the filename
            savedir (str)         : Save the resulting plot in this directory

        """
        wfplot = p.figure()
        # plot waveforms in a grid, always 3 colums
        lines = int(nwaveforms/3)
        if lines == 0 : lines = 1
        nwaveforms = lines*3

        #times = np.array([k for k in range(len(self.channel_data[ch][0]))])
        for i in tqdm.tqdm(range(nwaveforms), total=nwaveforms):
            ax = wfplot.add_subplot(lines, 3, i+1)
            volts = self.channel_data[ch][i]
            # FIXME - temporary workaround
            times = np.array([k for k in range(len(volts))])
            ax = plot_waveform(ax, times, volts)     
            # plot the axes only on the leftmost plots
            # and the downmost
            if (i+1 - 3) == 0:
                ax.set_ylabel("d. chan.")
            if (i+1 - nwaveforms) < 4:
                ax.set_xlabel("time [$\mu$s]")
        stripname = get_stripname(ch) 
        if not prefix:
            savename = f'waveforms-{stripname}.png' 
        else:
            savename = f'waveforms-{prefix}-{stripname}.png'
        savename = os.path.join(savedir, savename)
        wfplot.savefig(savename)
        return savename 

    def analyze(self, channel, save_shp_file=False):
        """
        Applyt the gaussian shaping algorithm on the waveform data.
        
        Args:
            channel (int)        : Select the channel

        Keyword Args:
            save_shp_file (bool) : Save a file with the shaping times

        """
        ptime_energies = dict()
        if not self.recordlengths:
            self.get_recordlengths()

        data = copy(self.channel_data[channel])
        for ptime in tqdm.tqdm(self.peakingtime_sequence):
            #logger.info(f"Crunching numbers for peaking time {ptime}")

            if self.adjust_shaper_order_dynamically:
                if 500 < ptime <= 1000:
                    order = 3
                elif ptime <= 500:
                    order = 2
                else:
                    order = 7
            else:
                order = self.order
            if self.use_simple_trapezoid_shaper:
                #shaper = TrapezoidalFilter(ptime = ptime, recordlength = self.recordlengths[channel])
                shaper = sh.TrapezoidalFilter(ptime, 1000, self.recordlengths[channel])
            else:
                shaper = sh.GaussShaper(ptime, order=order)
            #if self.njobs > 1:
            energies = np.array([energy for energy in self.ppexecutor.map(shaper.shape_it, data, chunksize=50)])
            #else:
            #energies = np.array([shaper.shaper_it(wf) for wf in data])
            ptime_energies[ptime] = energies 

        #if self.njobs > 1:
        #    for future in fut.as_completed(future_to_fname):
        #        ch, data = future.result()    
            
        return ptime_energies 




import hjson
import numpy as np
import pylab as p
import time 
from . import _pyCaenN6725 as _cn

import hepbasestack as hep
import hepbasestack.layout as lo
import hepbasestack.visual as vs
hep.visual.set_style_default()

from HErmes.visual.canvases import YStackedCanvas
import HErmes.visual.plotting as plt

landscape_half_figure_factory= lambda : p.figure(dpi=150, figsize=lo.FIGSIZE_A4_LANDSCAPE_HALF_HEIGHT)

class CaenN6725(object):
    """
    A digitizer from CAEN. Model name N6725.
    """
    vprobe1_to_str = {\
    _cn.DPPVirtualProbe1.Input  : 'Input',\
    _cn.DPPVirtualProbe1.Delta  : 'Delta',\
    _cn.DPPVirtualProbe1.Delta2 : 'Delta2',\
    _cn.DPPVirtualProbe1.Trapezoid : 'Trapezoid'}

    vprobe2_to_str = {\
    _cn.DPPVirtualProbe2.Input             : 'Input',\
    _cn.DPPVirtualProbe2.TrapezoidReduced  : 'TrapezoidReduced',\
    _cn.DPPVirtualProbe2.Baseline          : 'Baseline',\
    _cn.DPPVirtualProbe2.Threshold         : 'Threshold',\
    _cn.DPPVirtualProbe2.NONE              : 'None'}

    dprobe1_to_str = {\
    _cn.DPPDigitalProbe1.TRGWin            :  'TRGWin',\
    _cn.DPPDigitalProbe1.Armed             :  'Armed',\
    _cn.DPPDigitalProbe1.PkRun             :  'PkRun',\
    _cn.DPPDigitalProbe1.Peaking           :  'Peaking',\
    _cn.DPPDigitalProbe1.CoincWin          :  'CoincWin',\
    _cn.DPPDigitalProbe1.TRGHoldoff        :  'TRGHoldoff',\
    _cn.DPPDigitalProbe1.ACQVeto           :  'ACQVeto',\
    _cn.DPPDigitalProbe1.BFMVeto           :  'BFMVeto',\
    _cn.DPPDigitalProbe1.ExtTRG            :  'ExtTRG',\
    _cn.DPPDigitalProbe1.Busy              :  'Busy',\
    _cn.DPPDigitalProbe1.PrgVeto           :  'PrgVeto',\
    _cn.DPPDigitalProbe1.PileUp            :  'PileUp',\
    _cn.DPPDigitalProbe1.BLFreeze          :  'BLFreeze'}
    
    dprobe2_to_str = {\
    _cn.DPPDigitalProbe2.Trigger           :  'Trigger'}

    @property
    def samplingrate(self):
        return self.digitizer.get_current_sampling_rate()

    @staticmethod
    def parse_configfile(configfile):
        """
        Use hjson to parse a configfile
        with comments
        """

        config = hjson.load(open(configfile))
        if 'CaenN6725' in config.keys():
            config = config['CaenN6725']
        return config
        
    def __init__(self,
                 config,
                 has_dpp_pha_firmware=True,
                 shaping_time=None,
                 logger=None,
                 loglevel=30):
        """
        Instantiate a new Caen digitizer object. This allows to interace with 
        the CaenN6725 digitizer. 

        Args:
            config (dict) : This is the parsed result of a complex configfile
                            typically stored in .json format. The config shall 
                            contain parameters such as sample length, pulse 
                            polarity and so on.
                            In case the digitizer has a DPP_PHA firmware, the 
                            config shall contain the DPP_PHA parameters as well
        Keyword args:
            has_dpp_pha_firmware (bool) : True if the digitizer has the DPP_PHA firware
                                          (has to be purchased seperatly)

            shaping_time (int)          : In case the digitizer has a pulse shaping firmware
                                          the shaping time can already be set here.
            logger (logging.logger)     : A logging instance
            loglevel (int)              : log severity. 10 - debug, 20 - info, 30 - warn 
        """
        if has_dpp_pha_firmware:
            self.digitizer = _cn.CaenN6725DPPPHA()
        else:
            self.digitizer = _cn.CaenN6725WF()
        self.recordlength = None
        self.config = config
        self.trigger_thresholds = dict()
        self.dc_offsets = dict()
        self.dynamic_range = list()
        self.has_dpp_pha_firmware = has_dpp_pha_firmware
        self.acquisition_started = False
        self.shaping_time = shaping_time
        
        if logger is None:
            self.logger = hep.logger.get_logger(loglevel)
        else:
            self.logger = logger
        self.setup()
        self.logger.info("Caen N6725 initialization finished!")

    def __del__(self):
        if self.acquisition_started:
            self.digitizer.end_acquisition()
        del self.digitizer

    def get_handle(self):
        return self.digitizer.get_handle()

    def set_handle(self, handle):
        self.digitizer.set_handle(handle)

    @staticmethod
    def baseline_offset_percent_to_val(percent):
        """
        Convert the value of the baseline offset from 
        percentage to something the digitizer can understand
        """
        val = int((((2**16) -1)/100)*percent)
        return val

    def get_board_info(self):
        """
        Show information about digitizer model and firmware version
        """
        binfo = self.digitizer.get_board_info()
        self.logger.info(f"Found board with ROC FW: {binfo.get_model_roc_firmware_rel()} AMC FW: {binfo.get_amc_firmware_rel()} family code: {binfo.family_code} and serial number {binfo.serial_number}")
        return self.digitizer.get_board_info()

    def set_vprobe1(self, vprobe1):
        self.digitizer.set_vprobe1(vprobe1)

    def set_vprobe2(self, vprobe2):
        self.digitizer.set_vprobe2(vprobe2)

    def set_dprobe1(self, vprobe1):
        self.digitizer.set_vprobe1(vprobe1)

    def set_dprobe2(self, vprobe2):
        self.digitizer.set_vprobe2(vprobe2)
    
    def connect(self):
        self.digitizer.connect()


    def setup(self):
        config = self.config
        if not config:
            self.logger.warning("No config found, not changing any settings!")
            return None


        # active channels, digitizer dynamic range as well as baseline offsets
        # are set separately
        self.logger.debug(f'Configuring with {config}')

        # get the active channels
        active_digitizer_channels = config['active-channels']
        
        # baseline offset for active digitizier channels
        digitizer_baseline_offset = config['baseline-offset']
        assert len(digitizer_baseline_offset) == len(active_digitizer_channels), "The number of active channels does not match the number of baseline offsets given!"

        # the trigger thresholds for the active digitizer channels
        trigger_thresholds = config['trigger-threshold']
        assert len(trigger_thresholds) == len(active_digitizer_channels), "The number of active channels does not match the number of given trigger thresholds!"

        # dynamic range 
        digitizer_dynamic_range = config['dynamic-range']
        assert (digitizer_dynamic_range == "2VPP") or (digitizer_dynamic_range == "05VPP"), "Dynamic range has to be either 2VPP (2 volt peak-peak) or 05VPP (0.5 volt peak-peak"
        if digitizer_dynamic_range == "2VPP":
            digitizer_dynamic_range = _cn.DynamicRange.VPP2
            self.dynamic_range = [-1,1]
        if digitizer_dynamic_range == "05VPP":
            digitizer_dynamic_range = _cn.DynamicRange.VPP05
            self.dynamic_range = [-0.5, 0.5]
        
        # as an example, for now just take data with the digitzer
        digi_pars = self.extract_digitizer_parameters(config)
    
        #self.digitizer = _cn.CaenN6725(digi_pars)
        self.digitizer.connect()
        self.digitizer.configure(digi_pars)

        bf = self.digitizer.get_board_info()
        self.logger.debug(f'Connected to digitizer model {bf.get_model_name()}, roc firmware {bf.get_model_roc_firmware_rel()},  amc firmware {bf.get_amc_firmware_rel()}')

        # set input dynmaic range for all channels
        self.digitizer.set_input_dynamic_range(digitizer_dynamic_range)

        # set baseline offset
        # baseline offset has to go extra, for some reason CAEN treats this separatly
        for i,ch in enumerate(active_digitizer_channels):
            offset = self.baseline_offset_percent_to_val(digitizer_baseline_offset[i]) 
            self.logger.debug(f"Calculated dc offset for channel {ch} of {offset}")
            self.digitizer.set_channel_dc_offset(ch,offset)

            if not self.has_dpp_pha_firmware:
                self.digitizer.set_channel_trigger_threshold(ch, trigger_thresholds[ch])
                
            self.dc_offsets[ch] = offset
            # configure each channel individually
            #threshold = 
            if self.has_dpp_pha_firmware:
                dpp_params = self.extract_dpp_pha_parameters(ch, config, offset) 
                self.digitizer.configure_channel(ch, dpp_params)

        for ch, val in enumerate(self.digitizer.get_temperatures()):
            self.logger.debug(f'Chan: {ch} -  {val}\N{DEGREE SIGN}C')
        self.logger.debug("Will calibrate the digitizer")
        self.digitizer.calibrate()
        self.digitizer.allocate_memory()
        self.logger.info("Digitizer set up!")
        return 

    def extract_digitizer_parameters(self,config):
        """
        Extract the general configuration parameters
        for the digitizer from a config file 
        read by json
        """
        pars                    = _cn.DigitizerParams()
        # the enums we fix      forcn
        pars.LinkType           = _cn.ConnectionType.USB
        pars.IOlev              = _cn.IOLevel.NIM
        pars.PulsePolarity      = _cn.PulsePolarity.Positive
        # Fix this in the C++ code
        # pars.DPPAcqMode       = _cn.DPPAcqMode.Mixed
        active_channels         = config['active-channels']
        channelmask  = 0
        for ch in active_channels:
            channelmask += (1 << ch)
        pars.ChannelMask        = channelmask
        pars.VMEBaseAddress     = config['VMEBaseAddress']
        pars.RecordLength       = config['RecordLength']
        # also stroe in own field
        self.recordlength       = config['RecordLength']
        # this trigger threshold has NOTHING to do with the 
        # DPP-PHA trigger threshold, that is configured 
        # with the dpp-pha parameters later on 
        self.trigger_thresholds = config['trigger-threshold'] 
        pars.EventAggr          = config['EventAggr']
        if 'PostTriggerPercent' in config:
            pars.PostTriggerPercent = config['PostTriggerPercent']
        else:
            pars.PostTriggerPercent = 90
        return pars

    def get_temperatures(self):
        """
        Returns the temperatures of the digitizer electronics. One temperature 
        per channel
        """
        return self.digitizer.get_temperatures()

    def extract_dpp_pha_parameters(self, channel, config,\
                                   dc_offset,
                                   dynamic_range=[-0.5, 0.5]):
        """
        Extract the config parameters for the DPP-PHA
        algorithm from a config file 
        read by json
        
        Args:
            channel (int)    : channel number (0-7)
            config (dict)    : parsed config file
            dc_offset (list) : the set channel dc offset. This is necessary to calculate the trigger leve
                               the channel_dc need to be given in digitizer bins
        Keyword Args:
            dynamic_range (tuple) : V_pp dynamic range [min, max] in Volt

        """
        pars = _cn.DPPPHAParams()
        nchan = 8
        channel_key     = "ch" + str(channel)
        #self.logger.debug(f"Got trigger threshold : {config[channel_key]['trigger-threshold']}")
        trigger_threshold = config[channel_key]['trigger-threshold'] #in minivolts, neeed to convert to bin
        ## resolution is 14 bit
        #fsr = dynamic_range[1] - dynamic_range[0]
        #lsb = fsr/2**14
        #trigger_threshold = 1e-3*trigger_threshold # convert to volt
        #trigger_threshold = int(trigger_threshold/lsb)
        #self.logger.debug(f"Calculated value for {trigger_threshold}")
        #self.logger.debug(f"Found a dc offset of {dc_offset}")
        ##trigger_threshold += dc_offset/4 
        #trigger_threshold = int(trigger_threshold)
        self.logger.debug(f"This means we have a trigger threshold after applying the dc offset of {trigger_threshold}")
        #self.trigger_thresholds[channel] = trigger_threshold

        # check if the trapezoid is short enough to fit into the record length
        # reminder, the trapezoid rise time is our SHAPINGTIME
        if self.shaping_time is None:
            trrt  = config[channel_key]['trapezoid-rise-time']
        else:
            self.logger.warning('Warning, overwriting shaping time from config setting with {self.shaping_time}')
            trrt = self.shaping_time
        trft  = config[channel_key]['trapezoid-flat-top']
        trpho = config[channel_key]['peak-holdoff']
        pkrun = trrt + trft + trpho # in ns
        eventlength  = self.recordlength*(1/self.samplingrate)*1e9 # ns
        if pkrun >= eventlength: 
            raise ValueError(f"Trapezoid run time {pkrun} is longer than recordlength {self.recordlength}.Tof fix it, change the values of trapezoid-rise-time, trapezoid-flat-top, peak-holdoff.")
        self.logger.debug(f"We have an eventlength of {eventlength} and trapezoid run time of {pkrun}")
        pars.thr        = [trigger_threshold                                    ]*nchan 
        pars.k          = [trrt                                                 ]*nchan                     
        #pars.k          = [config[channel_key]['trapezoid-rise-time']           ]*nchan 
        pars.m          = [config[channel_key]['trapezoid-flat-top']            ]*nchan 
        pars.M          = [config[channel_key]['decay-time-constant']           ]*nchan 
        pars.ftd        = [config[channel_key]['flat-top-delay']                ]*nchan 
        pars.a          = [config[channel_key]['trigger-filter-smoothing-factor']]*nchan
        pars.b          = [config[channel_key]['input-signal-rise-time']        ]*nchan 
        pars.trgho      = [config[channel_key]['trigger-hold-off']              ]*nchan 
        pars.nsbl       = [config[channel_key]['n-samples']                     ]*nchan 
        pars.nspk       = [config[channel_key]['peak-mean']                     ]*nchan 
        pars.pkho       = [config[channel_key]['peak-holdoff']                  ]*nchan 
        pars.blho       = [config[channel_key]['baseline-holdoff']              ]*nchan
        pars.enf        = [config[channel_key]['energy-normalization-factor']   ]*nchan 
        pars.decimation = [config[channel_key]['decimation']                    ]*nchan 
        pars.dgain      = [config[channel_key]['decimation-gain']               ]*nchan 
        pars.otrej      = [config[channel_key]['otrej']                         ]*nchan 
        pars.trgwin     = [config[channel_key]['trigger-window']                ]*nchan 
        pars.twwdt      = [config[channel_key]['rise-time-validation-window']   ]*nchan
        return pars

    @staticmethod
    def guess_voltages(waveform, dynamic_range, offset):
        resolution = 2**14
        frs = dynamic_range[1] - dynamic_range[0]
        #offset = self.dc_offset[ch]
        #offset = (offset/4)*(frs/resolution)  + self.dynamic_range[0]
        if hasattr(waveform, '__iter__'):
            volts = np.array([w*(frs/resolution) + dynamic_range[0] for w in waveform]) 
        else:
            volts = waveform*(frs/resolution) + dynamic_range[0]
        return volts
    

    def to_volts(self,ch, waveform):
        """
        Convert waveform from digitizer bins to volts

        """
        resolution = 2**14
        frs = self.dynamic_range[1] - self.dynamic_range[0]
        #offset = self.dc_offset[ch]
        #offset = (offset/4)*(frs/resolution)  + self.dynamic_range[0]
        if hasattr(waveform, '__iter__'):
            volts = np.array([w*(frs/resolution) + self.dynamic_range[0] for w in waveform]) 
        else:
            volts = waveform*(frs/resolution) + self.dynamic_range[0]
        return volts


    def get_times(self, digital_trace=False):
        """
        Return the time binning in microseconds

        """
        binsize = 1/self.samplingrate
        if digital_trace:
            # FIXME: binsize always full sampling rate
            binsize = 1/250e6
        bins = np.array([k*binsize for k in range(self.recordlength)])
        return bins

    def run_digitizer(self,\
                      seconds,\
                      rootfilename=None,\
                      scope_mode=False,\
                      read_waveforms=False):
        """
        For the interactive use in an ipython notebook
    
        Args:
            seconds   (int)       : runtime in seconds

        Keyword Args:
            rootfilename  (str)   : filename of the output root file
            read_waveforms (bool) : sawe waveform data to the output root file
        """
        self.logger.info('Setting up digitizer...')
        if read_waveforms and self.has_dpp_pha_firmware:
            # for the other firmware, it can ONLY read out waveforms
            self.digitizer.enable_waveform_decoding()
            #self.digitizer.set_dprobe2(_cn.DPPDigitalProbe2.Trigger)
        
        if rootfilename is not None:
            self.digitizer.set_rootfilename(rootfilename)
        # run calibration before readout
        self.digitizer.calibrate()
        self.logger.info("Starting run")
        self.logger.info(f'This digitizer does have DPP PHA fw {self.has_dpp_pha_firmware}')
        self.digitizer.start_acquisition()
        self.acquisition_started = True
        if not self.has_dpp_pha_firmware:
            self.digitizer.readout_and_save(seconds)
        else:
            self.digitizer.continuous_readout(seconds)
        self.digitizer.end_acquisition()
        self.logger.info(f"We saw {self.digitizer.get_n_events_tot()} events!")
        return

    def live_view(self,\
                  seconds,
                  fill_histogram=False):
        """
        Return events for seconds. Use for oscilloscope live view
        
        Args:
            seconds (int)         : Stop acquisition after seconds.
        
        Keyword Args:
            fill_histogram (bool) : Fill the internal energy histogram
        """

        self.digitizer.calibrate()
        self.digitizer.start_acquisition()
        if self.has_dpp_pha_firmware:
            self.digitizer.clear_energy_histogram()
        delta_t = 0
        last = time.monotonic()
        while delta_t <= seconds:
            if self.has_dpp_pha_firmware:
                events = self.digitizer.read_data(fill_histogram)
                events = []
                for ch in range(8):
                    events.append(self.digitizer.get_last_waveform(ch))
                time.sleep(1)
                detla_t += 1 
            else:
                events = self.digitizer.readout_and_return()
            delta_t += time.monotonic() - last 
            last = time.monotonic()
            yield events
        self.digitizer.end_acquisition()
        self.logger.info(f"We saw {self.digitizer.get_n_events_tot()} events!")


    def init_scope(self):
        self.canvas = YStackedCanvas(subplot_yheights=(0.1, 0.1, 0.2, 0.5),
                                padding=(0.15, 0.05, 0.0, 0.1 ),\
                                space_between_plots=0,\
                                figsize="auto",\
                                figure_factory=None)


    def oscilloscope(self,\
                     maxtime=np.inf,\
                     filename="test.png",\
                     trace1 = _cn.DPPVirtualProbe1.Input,\
                     trace2 = _cn.DPPVirtualProbe2.TrapezoidReduced,\
                     dtrace1 = _cn.DPPDigitalProbe1.Peaking,\
                     dtrace2 = _cn.DPPDigitalProbe2.Trigger):
        """
        Show waveforms in oscilloscope mode 

        Keyword Args:
            maxtime (float)    : Maximum time the plot is active (in sec)
        """
        time_since_running = 0
        self.digitizer.enable_waveform_decoding()
        #self.canvas = YStackedCanvas(subplot_yheights=(0.1, 0.1, 0.2, 0.5),
        #                        padding=(0.15, 0.05, 0.0, 0.1 ),\
        #                        space_between_plots=0,\
        #                        figsize="auto",\
        #                        figure_factory=None)
        # initialize figure
        xs_analog  = self.get_times()*1e6 # times is in ns 
        xs_digital = self.get_times(digital_trace=True)*1e6
        # FIXME
        #xs = xs_digital # find out in what conditions the smapling rate is reduced
        #print (len (xs_analog))
        #print (len (xs_digital))
        xs = xs_analog
        #ax.set_xlabel("Time [$\mu$s]")
        #ax.set_ylabel("Voltage mV")
        p.ion()
        #ax.set_xlim(left=min(xs), right=max(xs))

        # trace 1 of the scope is the fast timing filter
        trace1_axes  = self.canvas.select_axes(0)
        trace2_axes  = self.canvas.select_axes(1)
        dtrace1_axes = self.canvas.select_axes(2)
        dtrace2_axes = self.canvas.select_axes(3)
        trace1_axes .clear() 
        trace2_axes .clear() 
        dtrace1_axes.clear() 
        dtrace2_axes.clear() 
        all_axes = trace1_axes, trace2_axes, dtrace1_axes, dtrace2_axes

        for ax in all_axes:
            ax.set_xlim(left=min(xs), right=max(xs))
        
        trace1_axes.set_xlabel("Time [$\mu$s]")

        if trace1 == _cn.DPPVirtualProbe1.Input:
            trace1_axes.set_ylabel("Voltage mV")

        # waveform (input) plot
        trace1_plot  = trace1_axes.plot(range(0), color="b", lw=1.2)[0]
        trace2_plot  = trace2_axes.plot(range(0), color="r", lw=1.2)[0]
        dtrace1_plot = dtrace1_axes.plot(range(0), color="k", lw=1.2)[0]
        dtrace2_plot = dtrace2_axes.plot(range(0), color="g", lw=1.2)[0]

        start_time = time.monotonic()
     
        #canvas.figure.show()
        #canvas.figure.canvas.draw()
        #canvas.figure.tight_layout()
   
        # set the traces -m
        self.digitizer.set_vprobe1(trace1)
        self.digitizer.set_vprobe2(trace2)
        self.digitizer.set_dprobe1(dtrace1)
        self.digitizer.set_dprobe2(dtrace2)

        #self.digitizer.allocate_memory()
        self.digitizer.start_acquisition()
        data = []
        while len(data) == 0:
            data = self.digitizer.read_data(0)
            if sum(self.digitizer.get_digital_trace2()) < 1:
                data = []

        self.digitizer.end_acquisition()

        ys_trace1  = self.digitizer.get_analog_trace1()
        ys_trace2  = self.digitizer.get_analog_trace2()
        ys_dtrace1 = self.digitizer.get_digital_trace1()
        ys_dtrace2 = self.digitizer.get_digital_trace2()
        energy     = self.digitizer.get_energy()

        print ("-----sneak peak waveforms-----")
        print ("xs",  xs[:10],  len(xs),  "timing bins")
        print ("tr1",  ys_trace1[:10],  len(ys_trace1),  self.vprobe1_to_str[trace1])
        print ("tr2",  ys_trace2[:10],  len(ys_trace2),  self.vprobe2_to_str[trace2])
        print ("dtr1", ys_dtrace1[:10], len(ys_dtrace1), self.dprobe1_to_str[dtrace1])
        print ("dtr2", ys_dtrace2[:10], len(ys_dtrace2), self.dprobe2_to_str[dtrace2])
        print ("energy", energy)
        print ("------------------------------")
        
        energy    = 1e3*self.to_volts(0, energy)
        plot_energy = None
        if trace1 == _cn.DPPVirtualProbe1.Input:
            ys_trace1  = self.to_volts(0,ys_trace1)
            ys_trace1 = 1e3*ys_trace1
            delta = abs(max(ys_trace1)) - abs(min(ys_trace1))
            bl_corrected = abs(ys_trace1) - abs(min(ys_trace1))*np.ones(len(ys_trace1))
            # calculate decay time, that is ln(2)*half signal height   
            try: 
                over_threshold = bl_corrected > delta/2
                half_time = xs[over_threshold][-1]
                self.logger.debug(f"Calculated decay time of {np.log(2)*half_time}")
            except Exception as e:
                self.logger.info("Halftime calculation failed, no pulse detected.")

            energy    = 1e3*self.to_volts(0, energy)
            plot_energy = 1
        #if trace2 == _cn.DPPVirtualProbe2.Input:
            #ys_trace2  = self.to_volts(0,ys_trace2)
            #ys_trace2 = 1e3*ys_trace2
            #plot_energy = 2
            #print (ys_trace1[:10])
        #print (ys_trace2[:10])
        #print (len(xs), len(ys_trace1), len(ys_trace2))
        #all_plots = trace1_plot, trace2_plot, dtrace1_plot, dtrace2_plot


        trace1_plot.set_xdata(xs)
        trace1_plot.set_ydata(ys_trace1)
        #if len(ys_trace2):
        trace2_plot.set_xdata(xs)
        trace2_plot.set_ydata(ys_trace2)
        dtrace1_plot.set_xdata(xs)
        dtrace1_plot.set_ydata(ys_dtrace1)
        dtrace2_plot.set_xdata(xs)
        dtrace2_plot.set_ydata(ys_dtrace2)

        if plot_energy == 1:
            trace1_axes.hlines(min(xs), max(xs), energy)
        if plot_energy == 2:
            trace2_axes.hlines(min(xs), max(xs), energy)

        for ys, ax in [(ys_trace1, trace1_axes),\
                       (ys_trace2, trace2_axes),\
                       (ys_dtrace2, dtrace1_axes),\
                       (ys_dtrace2, dtrace2_axes)]:
            if len(ys):
                ymax = max(ys) + abs(max(ys)*0.2)
                ymin = min(ys) - abs(min(ys)*0.2)
                self.logger.debug (f"adjugsting range {ymin} - {ymax}")
                ax.set_ylim(bottom=ymin, top=ymax)
                ax.grid(True)

        #canvas.figure.tight_layout()

        vs.adjust_minor_ticks(trace1_axes, which = 'both')
        vs.adjust_minor_ticks(trace2_axes, which = 'y')
        #plt.adjust_minor_ticks(dtrace1_axes, which = 'y')
        #plt.adjust_minor_ticks(dtrace2_axes, which = 'y')
                

        trace1_axes.text(xs[int(0.7*len(xs))],0.8*max(ys_trace1),   self.vprobe1_to_str[trace1]) 
        trace2_axes.text(xs[int(0.7*len(xs))],0.8*max(ys_trace2),   self.vprobe2_to_str[trace2]) 
        dtrace1_axes.text(xs[int(0.7*len(xs))],0.8*max(ys_dtrace1), self.dprobe1_to_str[dtrace1]) 
        dtrace2_axes.text(xs[int(0.7*len(xs))],0.8*max(ys_dtrace2), self.dprobe2_to_str[dtrace2]) 

        # despine
        trace2_axes.spines['bottom'].set_visible(False)
        dtrace1_axes.spines['bottom'].set_visible(False)
        dtrace2_axes.spines['bottom'].set_visible(False)
        for tic in trace2_axes.xaxis.get_major_ticks():
            tic.tick1line.set_visible(False)
            #tic.tick1On = tic.tick2On = False
        for tic in dtrace1_axes.xaxis.get_major_ticks():
            tic.tick1line.set_visible(False)
            #tic.tick1On = tic.tick2On = False
        for tic in dtrace2_axes.xaxis.get_major_ticks():
            tic.tick1line.set_visible(False)
            #tic.tick1On = tic.tick2On = False


        self.canvas.figure.canvas.draw()
        self.canvas.figure.savefig(filename)
        #for i in range(6):
        #    time.sleep(5)
        #for k in data:
        #   for j in k:
        #       print (j.wavefor
        #ms)
        #while True:
        #    sec = time.monotonic() - start_time
        #    datamins, datamaxes = [],[]
        #    for ch, line_plot in enumerate(line_plots):
     
        #        temp = self.get_temperature(ch)
        #        secs, temps = line_plot.get_data()
        #        secs = np.append(secs,sec)
        #        temps = np.append(temps,temp)
        #        line_plot.set_ydata(temps)
        #        line_plot.set_xdata(secs)
        #        datamins.append(min(temps))
        #        datamaxes.append(max(temps))
        #        xmin = min(secs)
        #        xmax = max(secs)
        #        if ch == feedback_ch:
        #            feedback_temp = temp
     
     
        #    datamax = max(datamaxes)
        #    datamin = min(datamins)
        #    if len(secs) == 1:
        #        continue
     
        #    # avoid matplotlib warning
        #    if abs(datamin - datamax) < 1:
        #        datamin -= 1
        #        datamax += 1
     
        #    if abs(xmax - xmin) < 1:
        #        xmin -= 1
        #        xmax += 1
     
        #    # update the plot
        #    ax.set_xlim(left=xmin, right=xmax)
        #    ax.set_ylim(bottom=datamin, top=datamax)
     
        #    fig.tight_layout()
        #    fig.canvas.draw()
        #    time.sleep(1)
        #    time_since_running += 1
     
        #    if time_since_running > maxtime:
        #        return
     
                 
     




"""
Software package to connect with a N675 digitizer made by CAEN.
Supports:
 - waveform readout firmware
 - DPP-PHA firmware
"""

__version__ = '0.0.17'

from .CaenN6725 import CaenN6725

LOGLEVEL = 20




import numpy as np
from copy import deepcopy as copy
import time
#peaktime = 1.0 # micro sec
#flat = 1.0 # micro sec
#nramp = peaktime/4e-3 # 4 ns per sample
#nflat = flat/4e-3 
#ntot = 2*nramp+nflat
#
#def trapezoidal_filter(waveform,
#                       ptime = 1.0,\
#                       flat  = 1.0,\
#                       nramp = 1./4e-3,\
#                       ntot  = 2*(1/4e-3) + 1/4e-3):
class TrapezoidalFilter(object):

    def __init__(self,\
                 ptime = 1000,\
                 flat  = 1000,
                 recordlength = 50000):
        """
        Keyword Args:
            ptime  (float)         :
            flat   (float)         :
            recordlength (int)     : waveform recordlength for preparing of the shaper windows
        """
        self.ptime = ptime
        self.flat  = flat
        self.nflat = int(flat/4)
        self.nramp = int(ptime/4) 
        self.ntot  = int(2*self.nramp + self.nflat)
        i_s = np.arange(self.ntot,recordlength,1)
        j_s = np.arange(0, self.nramp, 1)
        self.windows  = np.array([[i - j for j in j_s] for i in i_s])

    def shape_it(self, waveform):
        """
        Apply Mengjiao's simple trapezoidal filter

        Args:
            waveform (np.ndarray)  :  input waveform, baseline corrected 

        """
        amp_filter = np.array([(waveform[index] - waveform[index - self.nramp - self.nflat]).sum()/self.nramp for index in self.windows], dtype=np.int16)

        #amp_filter = np.array([waveform[riseindex].sum() + waveform[downindex].sum() for (riseindex, downindex) in zip(k,l)])
        # l   = i_s - (np.ones(len(i_s))*self.nramp) + (np.ones(len(i_s))*self.nflat) - j_s
        #sum_rise = waveform[k].sum()
        #sum_down = waveform[l].sum()
        #amp_filter = sum_rise + sum_down
        energy = max(amp_filter)
        return energy
        ##for i in range(self.ntot, len(waveform)):

        #    #j = 0
        #    #while j < self.nramp:
        #    for j in range(self.nramp):
        #        thisindex1 = i-j 
        #        thisindex2 = i - (self.nramp + self.nflat) - j 
        #        sum_rise += waveform[thisindex1]/self.nramp
        #        sum_down -= waveform[thisindex2]/self.nramp
        #        #j += 1     

        #    ##j = 0
        #    ##while j < self.nramp:
        #    #for j in range(self.nramp):
        #    #    thisindex = i - (self.nramp + self.nflat) - j 
        #    #    sum_down -= waveform[thisindex]/self.nramp
        #    #    #j += 1
        #    ampfilter = sum_rise + sum_down
        #    if ampfilter > energy:
        #        energy = ampfilter
        #    i += 1
        print (energy)
        return energy

#        for(int j=ntot;j<waveform->size();j++)
#        {   
#          cout<<j<<" "<<(*waveform)[j]<<endl;
#            hraw->Fill(j, (*waveform)[j]-baseline);
#
#            double amp_filter=0.;
#            double sum_rise=0.;
#            double sum_down=0.;
#
#            for(int ii=0;ii<nramp;ii++)
#            {
#                sum_rise += ((*waveform)[j-ii]-baseline)*1/(nramp*1.0);
#            }
#            for(int jj=0; jj<nramp;jj++)
#            {
#                sum_down += ((*waveform)[j-(nramp+nflat)-jj]-baseline)*(-1/(nramp*1.0));
#            }
#            amp_filter = sum_rise + sum_down;
#
#            if(amp_filter>energy)
#            {
#                energy = amp_filter;  //energy calculation, need to be improved
#            }
#
#            hfilter->Fill(j, amp_filter);
#        }




import numpy as np
from scipy import signal

def gaussian_shaper(order,peaktime,dt=1E-9,pz=0.0,user_tf=None):

	# Based on Ohkawa 1976 
	#"Direct synthesis of the Gaussian filter for nuclear pulse amplifiers"

	if order == 1:
		tf = 2 * 1.0844
		poles = [complex(-1,0)]
	elif order == 2:
		tf = 9.734458e-01
		c = np.sqrt(np.sqrt(2) +2)*0.5
		poles = [c*complex(-1,np.sqrt(2) - 1),c*complex(-1,1 - np.sqrt(2))]
	elif order == 3:
		tf = 6.740357e-01
		poles = [complex(-1.2633573,0),
		         complex(-1.1490948,0.7864188),complex(-1.1490948,-0.7864188)]
	elif order == 4:
		tf = 5.106046e-01
		poles = [complex(-1.3553576,0.3277948),complex(-1.3553576,-0.3277948),
		         complex(-1.1810803,1.0603749),complex(-1.1810803,-1.0603749)]
	elif order == 5:
		tf = 4.267639e-01
		poles = [complex(-1.4766878,0),
		         complex(-1.4166647,0.5978596),complex(-1.4166647,-0.5978596),
					complex(-1.2036832,1.2994843),complex(-1.2036832,-1.2994843)]
	elif order == 6:
		tf = 3.737515e-01
		poles = [complex(-1.5601279,0.2686793),complex(-1.5601279,-0.2686793),
		         complex(-1.4613750,0.8329565),complex(-1.4613750,-0.8329565),
					complex(-1.2207388,1.5145343),complex(-1.2207388,-1.5145343)]
	elif order == 7:
		tf = 3.371212e-01
		poles = [complex(-1.6610245,0),
		         complex(-1.6229725,0.5007975),complex(-1.6229725,-0.5007975),
					complex(-1.4949993,1.0454546),complex(-1.4949993,-1.0454546),
					complex(-1.2344141,1.7113028),complex(-1.2344141,-1.7113028)]
	else:
		raise ValueError('only gaussian shaper orders between 1 and 7 are supported')

	#this gave roughly the right answer, but a separate script solved for the tf values above, which give extremely acurate peaking times
	#sigma = 2 * 1.0844 * peaktime * (1./order)
	if user_tf is not None:
		tf = user_tf
	sigma = tf * peaktime
	poles = np.array(poles)/sigma
	zeros = np.array([-pz])
	dzpk = signal.bilinear_zpk(zeros,poles,1,1./dt)
	return dzpk,(zeros,poles)

def crrc_shaper(order,peaktime,dt=1E-9,pz=0.0):
	rc = peaktime/order
	poles = np.full(order + 1,-1./rc)
	zeros = np.array([-pz])
	dzpk = signal.bilinear_zpk(zeros,poles,1,1./dt)
	return dzpk,(zeros,poles)
	
def shaper(shape,order,peaktime,dt=1E-9,pz=0.0,normalize=True,return_zpk=False,bipolar=False):
	if shape == 'gaussian':
		f = gaussian_shaper
	elif shape == 'crrc':
		f = crrc_shaper
	else:
		raise ValueError('specified shaper shape not supported')
	dzpk,azpk = f(order,peaktime,dt=dt,pz=pz)
	zeros,poles,gain = dzpk
	if bipolar:
		zeros = np.append(zeros,[1])

	if normalize:
		sos = signal.zpk2sos(zeros,poles,1)
		t = np.arange(-1*peaktime,2*peaktime,dt)
		tail = np.zeros_like(t)
		tail[t>0] = np.exp(-t[t>0] * pz)
		ytail = signal.sosfilt(sos,tail)
		gain = 1./np.max(ytail)
		sos = signal.zpk2sos(zeros,poles,gain)
		dzpk = zeros,poles,gain
	else:
		sos = signal.zpk2sos(*dzpk)

	if return_zpk:
		return sos,dzpk,azpk
	else:
		return sos

#class Shaper:
#	shapes = ('crrc','gaussian','trapz')
#
#	shape: one of ('crrc','gaussian','trapezoid')
#	peaktime: the peaking time in seconds.  If trapezoid, then a 3-tuple of integers specifiying number of samples in rising edge, flat top, and falling edge.
#	order: the order (number of poles), default = 5. For shape = "gaussian", order must be <= 7.  For "crrc", it can be anything.
#	dt: the sample period in seconds, default = 32E-9 (32 nanoseconds)
#	tau: the decay time of the input pulse, default is 0 (step input).  If the input pulses are tail pulses, set tau = tail pulse decay time for pole zero correction.
#
#	def __init__(self,shape,peaktime,order=5,dt=32E-9,tau=0,trapz_causal=True)
#		self.peaktime = peaktime
#		self.dt = dt
#		self.tau = tau
#		if shape in shapes:
#			self.shape = shape
#		else:
#			raise ValueError("shape must be one of " + str(shapes))
#		if shape == 'trapezoid':
#			self.times = [8,8,8]
#			try:
#				self.times[0],self.times[1],self.times[2] = [int(t) for t in peaktime]
#			except Exception as E:
#				print("when shape = trapezoid, peaktime must be a 3-tuple of integers")
#				raise E
#			rise,flat,fall = self.times
#			self.k = np.concatenate([np.ones(rise)/rise,np.zeros(flat),-np.ones(fall)/fall])
#			if trapz_causal:
#				self.k = np.concatenate([np.zeros(len(self.k)),self.k])
#		if shape == 'gaussian':
#			if order > 7:
#				raise ValueError("maximum order for gaussian shaper is 7")
#			self.k = shaper("gaussian",order,peaktime,dt=dt,pz=1./tau)



"""
Gaussian shaper class primarily for the use with 
dactyos data.
"""
from scipy.signal import sosfilt

from .shapers import shaper

class GaussShaper(object):
    """
    A wrapper for Alex shaper software.

    """
    def __init__(self, peaktime, order=4,\
                 dt=4e-9, decay_time=80e-6):
        """
        
        Args:
            peaktime (float)  : peaking time in nanosecondd

        """
        # remember to convert everything to seconds
        self.sos = shaper("gaussian",order,1e-9*peaktime,dt=dt,pz=1./decay_time)

    def shape_it(self, tailpulses):
        #sos = shaper("gaussian",self.order,self.peaktime,dt=self.dt,pz=1./self.decay_time)
        #pulses = copy(tailpulses)
        #tailpulses_cr, bl = baseline_correction(pulses, nsamples=1024)
        y = sosfilt(self.sos,tailpulses)
        return max(y)






from .gauss_shaper import GaussShaper

WAVEFORMTYPE='ARRAY'
try:
    from _trapezoidal_shaper import TrapezoidalFilter
    WAVEFORMTYPE='LIST'
except ImportError:
    print ("WARNING, can not compile c++ extension for TrapezoidalFilter, Trapezoidal filter will be SLOW!")
    from .trapezoidal_shaper import TrapezoidalFilter
